/* 
 * kpf.c - parse OSCAR files for information. 
 *
 * Copyright 1993, by Kelvin J. Hill. 
 *
 */

#include <stdio.h>
#include <string.h>
#include <varargs.h>
#include <ctype.h>

#define	KPF_DATA "/BBS/KPF/kpf_data"

extern char    *optarg;
extern int      optind;

char           *
dgets(char *, int, FILE *);

char           *
stristr(char *, char *);

int             D_Ind = 0;	/* detailed entry */
int             K_Ind = 0;	/* keyword */
int             M_Ind = 0;	/* model type */
int             P_Ind = 0;	/* product name */
int             R_Ind = 0;	/* release number */
int             S_Ind = 0;	/* summary only */
int             V_Ind = 0;	/* product version */

int             Lines = 0;	/* line count */

char            star[7] = "";	/* KPF (star) number */
char            keyword[5][21]  = {"","","","",""};	/* keyword buffer */
char            model[5][10] = {"","","","",""};	/* model type */
char            product[10] = "";	/* product type */
char            release[10] = "";	/* release */
char            version[10] = "";	/* version */

#define BUFLEN 256
char            buffer[BUFLEN];

char            summary_1[BUFLEN];
char            summary_2[BUFLEN];

FILE           *kpf;

main(argc, argv)
	int             argc;
	char          **argv;
{
	int             c, i;

	if (argc < 2) {
		usage(argv[0]);
		exit(1);
	}
	while ((c = getopt(argc, argv, "d:k:m:p:r:sv:")) != EOF) {
		switch (c) {
		case 'd':
			D_Ind = 1;
			if (strlen(optarg) != 6) {
				dprintf("KPF number must be 6 characters long.\n");
				exit(1);
			}
			strcpy(star, optarg);
			break;

		case 'k':
			if(K_Ind == 5) {
                                dprintf("Too many keywords requested. First 5 only will be used.\n");
                                break;
                        }

			strncpy(keyword[K_Ind], optarg, 20);
			K_Ind += 1;
			break;

		case 'm':
			if(M_Ind == 5) {
                                dprintf("Too many models requested. First 5 only will be used.\n");
                                break;
                        }

			strncpy(model[M_Ind], optarg, 9);
			M_Ind += 1;
			break;

		case 'p':
			P_Ind = 1;
			strncpy(product, optarg, 9);
			break;

		case 'r':
			R_Ind = 1;
			strncpy(release, optarg, 9);
			break;

		case 's':
			S_Ind = 1;
			break;

		case 'v':
			V_Ind = 1;
			strncpy(version, optarg, 9);
			break;
		}
	}

	if (D_Ind & S_Ind) {
		dprintf("Only one option may be chosen.\n");
		exit(1);
	}
	if (!D_Ind & !S_Ind) {
		usage(argv[0]);
		exit(1);
	}
	if (S_Ind) {
		dprintf("Displaying Summary KPF Information for the following selections.\n");
		dprintf("Model(s): %s %s %s %s %s, O/S Release: %s, Product: %s, Version: %s\n",
			(M_Ind > 0) ? model[0] : "all",
			(M_Ind > 1) ? model[1] : "",
			(M_Ind > 2) ? model[2] : "",
			(M_Ind > 3) ? model[3] : "",
			(M_Ind > 4) ? model[4] : "",
			R_Ind ? release : "all",
			P_Ind ? product : "all",
			V_Ind ? version : "all");

		if(K_Ind) {
			dprintf("Keyword(s): %s %s %s %s %s\n",
				keyword[0], keyword[1], keyword[2], keyword[3], keyword[4]);
		}

		dprintf("\n");

		if ((kpf = fopen(KPF_DATA, "r")) == (FILE *) NULL) {
			perror(KPF_DATA);
			exit(2);
		}
		while (fgets(buffer, BUFLEN, kpf) != (char *) NULL) {
			if (!strncmp(buffer, "*******PS", 9) && vet()) {
				if (prepare_summary()) {
					dprintf("KPF: %s\n", star);
					dprintf("%s", summary_1);
					dprintf("%s", summary_2);
					dprintf("\n");
				}
			}
		}

		setraw();
		printf("*** Depress a key to continue ........ ");
		getchar();
		restore();
		printf("\n");
		fclose(kpf);
		exit(0);
	}
	if (D_Ind) {
		dprintf("Displaying Detailed KPF Information for %s\n", star);
		dprintf("\n");

		if ((kpf = fopen(KPF_DATA, "r")) == (FILE *) NULL) {
			perror(KPF_DATA);
			exit(2);
		}
		while (fgets(buffer, BUFLEN, kpf) != NULL) {
			if (!strncmp(buffer, "*******PS", 9) && vet()) {
				dprintf("KPF: %s\n", star);
				display_summary();
				display_detail();
				dprintf("\n");
				display_reply();
				dprintf("\n");
				setraw();
				printf("*** Depress a key to continue ........ ");
				getchar();
				restore();
				printf("\n");
				fclose(kpf);
				exit(0);
			}
		}
		printf("KPF number %s was not found on file.\n", star);
		setraw();
		printf("*** Depress a key to continue ........ ");
		getchar();
		restore();
		printf("\n");
		fclose(kpf);
		exit(1);
	}
}

vet()
{
	int             i = 0, j = 0;
	int             Match = 0;

	if (D_Ind && strncmp(star, &buffer[25], 6)) {
		return (0);
	}
	strncpy(star, &buffer[25], 6);
	star[6] = '\0';

	while (fgets(buffer, BUFLEN, kpf) != (char *) NULL) {
		if (!strncmp(buffer, "----------", 10)) {
			if (i++ == 1) {
				if (fgets(buffer, BUFLEN, kpf) != (char *) NULL && S_Ind) {
					for (j = 0; j < M_Ind; j++) {
						if (strlen(model[j]) && strncmp(model[j], &buffer[9], strlen(model[j]))) {
						} else {
							Match = 1;		/* Found at least one match */
						}
					}
					if (M_Ind && !Match)
						return(0);
				}
				if (fgets(buffer, BUFLEN, kpf) != (char *) NULL && S_Ind) {
					if (strlen(release) && strncmp(release, &buffer[58], strlen(release)))
						return (0);
				}
				if (fgets(buffer, BUFLEN, kpf) != (char *) NULL && S_Ind) {
					if (strlen(product) && strncmp(product, &buffer[9], strlen(product)))
						return (0);
					if (strlen(version) && strncmp(version, &buffer[34], strlen(version)))
						return (0);
				}
				return (1);
			}
		}
	}
}

prepare_summary()
{
	int i;

	summary_1[0] = '\0';
	summary_2[0] = '\0';

	summary_1[12] = '\0';
	summary_2[12] = '\0';

	while (fgets(buffer, BUFLEN, kpf) != (char *) NULL) {
		if (!strncmp(buffer, "----------", 10)) {
			if (dgets(buffer, BUFLEN, kpf) != (char *) NULL) {
				strcpy(summary_1, buffer);
			}
			if (dgets(buffer, BUFLEN, kpf) != (char *) NULL) {
				strcpy(summary_2, buffer);
			}
			if (!K_Ind) {
				return (1);
			}

			for (i = 0; i < K_Ind; i++) {
				if (stristr(&summary_1[12], keyword[i]) == (char *) NULL &&
				    stristr(&summary_2[12], keyword[i]) == (char *) NULL) {
					return (0);
				}
			}
			return (1);
		}
	}
}

display_summary()
{
	int             i = 0;

	while (fgets(buffer, BUFLEN, kpf) != (char *) NULL) {
		if (!strncmp(buffer, "----------", 10)) {
			for (i = 0; i < 2; i++) {
				if (dgets(buffer, BUFLEN, kpf) != (char *) NULL) {
					dprintf(buffer);
				}
			}
			return (0);
		}
	}
}

display_detail()
{
	while (dgets(buffer, BUFLEN, kpf) != (char *) NULL && strncmp(buffer, "----------", 10)) {
		dprintf(buffer);
	}
	return (0);
}

display_reply()
{
	int             i = 0;

	while (fgets(buffer, BUFLEN, kpf) != (char *) NULL && strncmp(buffer, "----------", 10)) {
	}

	while (fgets(buffer, BUFLEN, kpf) != (char *) NULL) {
		if (!strncmp(buffer, "----------", 10)) {
			while (dgets(buffer, BUFLEN, kpf) != (char *) NULL && strncmp(buffer, "==========", 10)) {
				dprintf(buffer);
				if (i++ == 1) {
					dprintf("\n");
				}
			}
			return (0);
		}
	}
}

usage(progname)
	char           *progname;
{
	dprintf("usage: %s [options]\n", progname);
	dprintf("       -s             summary off all kpf entries.\n");
	dprintf("       -d <kpf#>      detailed display of kpf entry.\n");
	dprintf("       -k <keyword>   select on summary keyword.\n");
	dprintf("       -m <model>     select on model type.\n");
	dprintf("       -r <release>   select on O/S release.\n");
	dprintf("       -p <product>   select on model type.\n");
	dprintf("       -v <version>   select on product version.\n");
}


dprintf(va_alist)
va_dcl
{
	va_list         args;
	char           *fmt;


	if (Lines == 22) {
		setraw();
		printf("*** Depress a key to continue ( control-k to quit ) ........ ");
		fflush(stdout);
		if (getchar() == 0x0b) {
			restore();
			fclose(kpf);
			exit(0);
		}
		restore();
		puts("\n");
		Lines = 0;
	}
	Lines++;

	va_start(args);
	fmt = (char *) va_arg(args, char *);
	vprintf(fmt, args);
	va_end(args());
	fflush(stdout);
}

char           *
dgets(buffer, buflen, kpf)
	char           *buffer;
	int             buflen;
	FILE           *kpf;
{
	char           *c;
	int             i;

	c = fgets(buffer, buflen, kpf);

	if (c == (char *) NULL) {
		return (c);
	}
	for (i = strlen(buffer) - 1; i >= 0; i--) {
		if (buffer[i] != ' ' && buffer[i] != '\n') {
			break;
		} else {
			buffer[i] = '\0';
		}
	}
	buffer[++i] = '\n';
	return (c);
}

char           *
stristr(target, source)
	char           *target;
	char           *source;
{
	char           *str1, *str2, *result, *c;

	if ((str1 = strdup(target)) == (char *) NULL) {
		return ((char *) NULL);
	}
	if ((str2 = strdup(source)) == (char *) NULL) {
		free(str1);
		return ((char *) NULL);
	}
	for (c = str1; *c != '\0'; c++) {
		if (isupper(*c)) {
			*c = tolower(*c);
		}
	}

	for (c = str2; *c != '\0'; c++) {
		if (isupper(*c)) {
			*c = tolower(*c);
		}
	}

	result = strstr(str1, str2);

	free(str1);
	free(str2);

	return (result);
}

#include <termio.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
static struct termio tbufsave;

syserr (msg)
	char   *msg;
{
	extern int errno, sys_nerr;
	extern char *sys_errlist[];
	fprintf (stderr, "ERROR:%s (%d", msg, errno);
	if (errno > 0 && errno < sys_nerr)
		fprintf (stderr, ";%s)\n", sys_errlist[errno]);
	else
		fprintf (stderr, ")\n");
	exit (1);
}

closedown()
{
	restore();
	exit(0);
}

setraw ()
{
	struct termio tbuf;
	if (ioctl (0, TCGETA, &tbuf) == -1)
		syserr ("ioctl setraw0");
	tbufsave = tbuf;
	tbuf.c_lflag = ISIG;
	tbuf.c_cc[VINTR] = 0x0b;/* Control K */
	tbuf.c_cc[4] = 1;
	if (ioctl (0, TCSETA, &tbuf) == -1)
		syserr ("ioctl setraw1");
	if (ioctl (1, TCSETA, &tbuf) == -1)
		syserr ("ioctl setraw2");
	signal(SIGINT, closedown);
}

restore ()
{
	if (ioctl (0, TCSETAF, &tbufsave) == -1)
		syserr ("ioctl restore");
}
